/*
i4004em Version 0.2.0 - Intel 4004 emulator
Copyright (C) 2001  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <curses.h>

/* ......ct */
#define TEST 0x01
#define TOS 0
#define MAXMEM 0xfff

int eofsw,pc;
int rega;
int reg0,reg1,reg2,reg3,reg4,reg5,reg6,reg7;
int reg8,reg9,reg10,reg11,reg12,reg13,reg14,reg15;
int regs,pc,startpc,src1,src2,cml;
int stack[4],ramport[4],ramstat[4];
int rominpt[16],romoutp[16];
int fcarry,ftest;
int pgmlen;
unsigned char *pgm,*pgmcurr,*pgmend;
char hxtbl[16];
unsigned char ramchar[256];

void putstx(pgm)
char *pgm;
   {
   fprintf(stderr,"Usage: %s <program>\n", pgm);
   exit(1);
   }

void initnc()
   {
   initscr();
   cbreak();
   noecho();
   nonl();
   intrflush(stdscr,FALSE);
   keypad(stdscr,TRUE);
   nodelay(stdscr,FALSE);
   clear();
   refresh();
   } /* initnc */

void disclaim()
   {
   int ch;
   mvaddstr(5,0,"i4004em version 0.1.0");
   mvaddstr(6,0,"Copyright (C) 2001  ");
   addstr("dondalah@ripco.com (Dondalah)");
   mvaddstr(7,0,"i4004em comes with ABSOLUTELY ");
   addstr("NO WARRANTY");
   mvaddstr(8,0,"For details, read the GNU ");
   addstr("Public License at");
   mvaddstr(9,0,"http://www.gnu.org/copyleft/gpl.html");
   mvaddstr(11,0,"Press any key to start.");
   curs_set(0);
   refresh();
   ch = getch();
   curs_set(1);
   refresh();
   } /* disclaim */

void dmpmem()
   {
   int hndl,wrlen,stat,mode;
   mode = S_IREAD | S_IWRITE;
   hndl = creat("tst.mem", mode);
   if (hndl < 0)
      {
      perror("Error creating tst.mem");
      exit(1);
      }
   wrlen = write(hndl,pgm,65536);
   if (wrlen < 0)
      {
      perror("Error writing tst.mem");
      exit(1);
      }
   stat = close(hndl);
   if (stat < 0)
      {
      perror("Error closing tst.mem");
      exit(1);
      }
   } /* dmpmem */

int loadseg(hndl)
int hndl;
   {
   int rdlen,len,pgmend;
   unsigned char buf[16];
   unsigned char *p;
   rdlen = read(hndl,buf,7);
   if (rdlen < 0)
      {
      fprintf(stderr,"Read error "
         "during program load\n");
      perror("Read error");
      exit(1);
      }
   else if (rdlen != 7)
      {
      fprintf(stderr,"Read error %02x "
         "during program load\n",
         rdlen);
      exit(1);
      }
   if (buf[0] == '\0') return(EOF);
   else if (buf[0] != 0x3f)
      {
      fprintf(stderr,"Invalid segment id %02x "
         "during program load\n",
         buf[0]);
      exit(1);
      }
   pc = (buf[2] << 8) | buf[1];
   p = pgm + pc;
   len = (buf[6] << 8) | buf[5];
   pgmend = pc + len;
   if (pgmend > 0xffff)
      {
      fprintf(stderr,"Read beyond program "
         "memory %06x during program load\n",
         pgmend);
      exit(1);
      }
   rdlen = read(hndl,p,len);
   if (rdlen < 0)
      {
      fprintf(stderr,"Read error "
         "during program load\n");
      perror("Read error");
      exit(1);
      }
   else if (rdlen != len)
      {
      fprintf(stderr,"Read error %02x "
         "during program load\n",
         rdlen);
      exit(1);
      }
   return(0);
   } /* loadseg */

void loadpgm(fname)
char *fname;
   {
   int i,rdlen,buflen,hndl,stat,pcl,pch;
   unsigned char *p;
   unsigned char buf[8];
   hndl = open(fname,O_RDONLY);
   if (hndl < 0)
      {
      fprintf(stderr,"Error opening %s\n",
         fname);
      perror("Open program error");
      exit(1);
      }
   buflen = MAXMEM + 1;
   rdlen = read(hndl,buf,2);
   if (rdlen < 0)
      {
      fprintf(stderr,"First read error "
         "during program load\n");
      perror("First read error");
      exit(1);
      }
   if (rdlen != 2)
      {
      fprintf(stderr,"Read error %02x "
         "during program load\n",
         rdlen);
      exit(1);
      }
   if (buf[0] != 0x89 || buf[1] != 0x14)
      {
      fprintf(stderr,"Invalid file id: "
         "%02x %02x\n",
         buf[0], buf[1]);
      exit(1);
      }
   stat = 0;
   while (!stat) stat = loadseg(hndl);
   startpc = pc = 0;
   pgmcurr = pgm + pc;
   pgmend = pgm + 0xffff;
   rega = reg0 = reg1 = 
      reg2 = reg3 = reg4 = reg5 =
      reg6 = reg7 = reg8 = reg9 =
      reg10 = reg11 = reg12 = reg13 =
      reg14 = reg15 = src1 = src2 = cml = 
      fcarry = ftest = 0;
   regs = TOS;
   for (i=0;i<4;i++)
      stack[i] = ramport[i] = ramstat[i] = 0;
   for (i=0;i<16;i++)
      rominpt[i] = romoutp[i] = 0;
   for (i=0;i<256;i++) ramchar[i] = 0;
   stat = close(hndl);
   if (stat < 0)
      {
      perror("Close program error");
      exit(1);
      }
   } /* loadpgm */

int getbyte()
   {
   int ch;
   unsigned char *p;
   char str[128];
   p = pgm + pc++;
   if (p >= pgmend)
      {
      sprintf(str,"Read beyond end of program\n");
      mvaddstr(20,0,str);
      sprintf(str,"pgmcurr %p "
         "pgmend %p pc %04x\n",
         pgmcurr, pgmend, pc);
      mvaddstr(21,0,str);
      refresh();
      dmpmem();
      endwin();
      exit(0);
      }
   ch = *p;
   return(ch);
   } /* getbyte */

int getval()
   {
   int ch,val,row,col,diff;
   char str[16];
   char *p,*q;
   mvaddstr(4,0,"Port value   (0-15): ");
   /*            0....+....1....+....2....+ */
   row = 4;
   col = 21;
   p = str;
   q = str+16;
   while (p < q)
      {
      diff = p - str;
      ch = getch();
      if (ch == 0x1b)
         {
         endwin();
         exit(0);
         } /* if quit */
      else if (ch == KEY_BACKSPACE)
         {
         if (p > str)
            {
            p--;
            mvdelch(row,--col);
            refresh();
            } /* if room */
         } /* if backspace */
      else if (ch == 8 || ch >= 0x7f)
         {
         if (p > str)
            {
            p--;
            mvdelch(row,--col);
            refresh();
            } /* if room */
         } /* if backspace */
      else if (ch == 13) break;
      else if (ch >= '0' && ch <= '9'
         && diff < 2)
         {
         *p++ = ch;
         mvaddch(row,col++,ch);
         refresh();
         } /* if valid digit */
      } /* read loop */
   *p = '\0'; 
   val = atoi(str);
   if (val < 0 || val > 15)
      {
      sprintf(str,"Invalid port "
         "value (not 0-15)\n");
      mvaddstr(10,0,str);
      refresh();
      ch = getch();
      if (ch == 0x1b)
         {
         endwin();
         exit(0);
         }
      return(-1);
      }
   return(val);
   } /* getval */

int getport()
   {
   int ch,port,row,col,diff;
   char str[16];
   char *p,*q;
   mvaddstr(3,0,"Port address (0-15): ");
   /*            0....+....1....+....2....+ */
   row = 3;
   col = 21;
   p = str;
   q = str+16;
   while (p < q)
      {
      diff = p - str;
      ch = getch();
      if (ch == 0x1b)
         {
         endwin();
         exit(0);
         } /* if quit */
      else if (ch == 8 || ch >= 0x7f)
         {
         if (p > str)
            {
            p--;
            mvdelch(row,--col);
            refresh();
            } /* if room */
         } /* if backspace */
      else if (ch == 13) break;
      else if (ch >= '0' && ch <= '9'
         && diff < 2)
         {
         *p++ = ch;
         mvaddch(row,col++,ch);
         refresh();
         } /* if valid digit */
      } /* read loop */
   *p = '\0'; 
   port = atoi(str);
   if (port < 0 || port > 15)
      {
      sprintf(str,"Invalid port "
         "address (not 0-15)\n");
      mvaddstr(10,0,str);
      refresh();
      ch = getch();
      if (ch == 0x1b)
         {
         endwin();
         exit(0);
         }
      return(-1);
      }
   return(port);
   } /* getport */

void readrom()
   {
   int ch,port,val;
   char str[32];
   port = getport();
   if (port == -1) return;
   val = getval();
   if (val == -1) return;
   rominpt[port] = val;
   } /* readrom */

void putnbl(ch)
int ch;
   {
   int lonbl;
   lonbl = ch & 15;
   addch( hxtbl[lonbl] );
   } /* putnbl */

void puthex(ch)
int ch;
   {
   int lonbl,hinbl;
   hinbl = ch >> 4;
   lonbl = ch & 15;
   addch( hxtbl[hinbl] );
   addch( hxtbl[lonbl] );
   } /* puthex */

void putdcml(reg)
int reg;
   {
   char str[16];
   sprintf(str,"%02d", reg&15);
   addstr(str);
   } /* putdcml reg */

void shwstat()
   {
   addstr("CT ");
   if (fcarry)
      {
      addch('C');
      }
   else
      {
      addch('.');
      }
   if (ftest)
      {
      addch('T');
      }
   else
      {
      addch('.');
      }
   addch(' ');
   } /* shwstat */

void shwromo()
   {
   int i;
   for (i=0;i<16;i++)
      {
      putdcml(romoutp[i]&15);
      addch(' ');
      if (i == 3 || i == 7 || i == 11)
         {
         addch(' ');
         addch(' ');
         } /* gap every 4th rom */
      } /* for each rom */
   } /* shwromo */

void shwromi()
   {
   int i;
   for (i=0;i<16;i++)
      {
      putdcml(rominpt[i]&15);
      addch(' ');
      if (i == 3 || i == 7 || i == 11)
         {
         addch(' ');
         addch(' ');
         } /* gap every 4th rom */
      } /* for each rom */
   } /* shwromi */

void dmpreg()
   {
   int ch;
   unsigned char *p;
   char str[128];
   p = pgm + pc;
   clear();
   sprintf(str,"(D)ump Program ROM  (I) ROM Input  "
      "(T)est  (R)eset  (Esc) quit");
   mvaddstr(0,0,str);
   sprintf(str,"(M) Display RAM  (G)o to address");
   mvaddstr(1,0,str);
   mvaddstr(15,0,"Cmd Line   ");
   putdcml(cml);
   mvaddstr(16,0,"SRC Pair   ");
   putdcml(src1);
   addch(' ');
   putdcml(src2);
   mvaddstr(17,0,"RAM Stat   ");
   putdcml(ramstat[0]&15);
   addch(' ');
   putdcml(ramstat[1]&15);
   addch(' ');
   putdcml(ramstat[2]&15);
   addch(' ');
   putdcml(ramstat[3]&15);
   addch(' ');
   mvaddstr(18,0,"RAM Ports  ");
   putdcml(ramport[0]&15);
   addch(' ');
   putdcml(ramport[1]&15);
   addch(' ');
   putdcml(ramport[2]&15);
   addch(' ');
   putdcml(ramport[3]&15);
   mvaddstr(19,0,"ROM Input  ");
   shwromi();
   mvaddstr(20,0,"ROM Output ");
   shwromo();
   mvaddstr(21,0," 0=");
   putdcml(reg0);
   addstr("  1=");
   putdcml(reg1);
   addstr("  2=");
   putdcml(reg2);
   addstr("  3=");
   putdcml(reg3);
   addstr("  4=");
   putdcml(reg4);
   addstr("  5=");
   putdcml(reg5);
   addstr("  6=");
   putdcml(reg6);
   addstr("  7=");
   putdcml(reg7);
   mvaddstr(22,0," 8=");
   putdcml(reg8);
   addstr("  9=");
   putdcml(reg9);
   addstr(" 10=");
   putdcml(reg10);
   addstr(" 11=");
   putdcml(reg11);
   addstr(" 12=");
   putdcml(reg12);
   addstr(" 13=");
   putdcml(reg13);
   addstr(" 14=");
   putdcml(reg14);
   addstr(" 15=");
   putdcml(reg15);
   mvaddstr(23,0," A=");
   putdcml(rega&0xff);
   addstr(" SP=");
   addch(regs+'0');
   addch(' ');
   putnbl((stack[regs]>>8)&15);
   puthex(stack[regs]&0xff);
   addstr(" PC=");
   putnbl((pc>>8)&15);
   puthex(pc&0xff);
   addch(' ');
   puthex(*p);
   addch(' ');
   puthex(*(p+1));
   addch(' ');
   shwstat();
   refresh();
   } /* dmpreg */

void shwmem()
   {
   int i,j,ch;
   unsigned char *p;
   char str[16];
   clear();
   mvaddstr(1,0,"RAM Memory");
   p = ramchar;
   for (i=0;i<16;i++)
      {
      sprintf(str,"%2d  ", i);
      mvaddstr(i+3,0,str);
      for (j=0;j<16;j++)
         {
         putdcml(*p++);
         addch(' ');
         if (j == 3 || j == 7 || j == 11)
            {
            addch(' ');
            addch(' ');
            } /* gap every 4 char */
         } /* row detail */
      } /* for each row */
   mvaddstr(23,0,"Press any key to return");
   refresh();
   ch = getch();
   } /* shwmem */

int getaddr()
   {
   int ch,val,row,col,diff;
   char str[16];
   char *p,*q;
   row = 10;
   col = 22;
   p = str;
   q = str+16;
   while (p < q)
      {
      diff = p - str;
      ch = getch();
      if (ch == 0x1b)
         {
         endwin();
         exit(0);
         } /* if quit */
      else if (ch == 8 || ch >= 0x7f)
         {
         if (p > str)
            {
            p--;
            mvdelch(row,--col);
            refresh();
            } /* if room */
         } /* if backspace */
      else if (ch == 13) break;
      else if (ch >= '0' && ch <= '9'
         && diff < 3)
         {
         *p++ = ch - '0';
         mvaddch(row,col++,ch);
         refresh();
         } /* if valid digit */
      else if (ch >= 'A' && ch <= 'F'
         && diff < 3)
         {
         *p++ = (ch - 'A') + 10;
         mvaddch(row,col++,ch);
         refresh();
         } /* if valid digit */
      else if (ch >= 'a' && ch <= 'f'
         && diff < 3)
         {
         *p++ = (ch - 'a') + 10;
         mvaddch(row,col++,ch);
         refresh();
         } /* if valid digit */
      } /* read loop */
   *p = 0;
   q = p;
   p = str;
   val = 0;
   while (p < q && val < 0x1000)
      {
      val = (val << 4) + *p++;
      } /* convert hex to binary */
   if (val < 0 || val > 0xfff)
      {
      sprintf(str,"Invalid address "
         "(not 0-fff)\n");
      mvaddstr(11,0,str);
      refresh();
      ch = getch();
      if (ch == 0x1b)
         {
         endwin();
         exit(0);
         }
      return(pc);
      }
   return(val);
   } /* getaddr */

void gotoloc()
   {
   int ch;
   char str[128];
   char *p,*q;
   mvaddstr(10,0,"Enter address in hex: ");
   /*             0....+....1....+....2....+....3 */
   refresh();
   pc = getaddr();
   } /* gotoloc */

void jcn0()
   {
   mvaddstr(3,0,"Unimplemented condition: 0");
   refresh();
   endwin();
   exit(0);
   } /* jcn0 */

void jcn1()
   {
   int addr;
   addr = getbyte() & 0xff;
   /* if (rega != 0 && !fcarry && ftest) */
   if (ftest)
      pc = (pc & 0xf00) | addr;
   } /* jcn1 */

void jcn2()
   {
   int addr;
   addr = getbyte() & 0xff;
   /* if (rega != 0 && fcarry && !ftest) */
   if (fcarry)
      pc = (pc & 0xf00) | addr;
   } /* jcn */

void jcn3()
   {
   mvaddstr(3,0,"Unimplemented condition: 3");
   refresh();
   endwin();
   exit(0);
   } /* jcn3 */

void jcn4()
   {
   int addr;
   addr = getbyte() & 0xff;
   /* if (rega == 0 && !fcarry && !ftest) */
   if (rega == 0)
      pc = (pc & 0xf00) | addr;
   } /* jcn4 */

void jcn5()
   {
   mvaddstr(3,0,"Unimplemented condition: 5");
   refresh();
   endwin();
   exit(0);
   } /* jcn5 */

void jcn6()
   {
   mvaddstr(3,0,"Unimplemented condition: 6");
   refresh();
   endwin();
   exit(0);
   } /* jcn6 */

void jcn7()
   {
   mvaddstr(3,0,"Unimplemented condition: 7");
   refresh();
   endwin();
   exit(0);
   } /* jcn7 */

void jcn8()
   {
   mvaddstr(3,0,"Unimplemented condition: 8");
   refresh();
   endwin();
   exit(0);
   } /* jcn8 */

void jcn9()
   {
   int addr;
   addr = getbyte() & 0xff;
   if (!ftest) pc = (pc & 0xf00) | addr;
   } /* jcn9 */

void jcn10()
   {
   int addr;
   addr = getbyte() & 0xff;
   if (!fcarry) pc = (pc & 0xf00) | addr;
   } /* jcn10 */

void jcn11()
   {
   mvaddstr(3,0,"Unimplemented condition: 11");
   refresh();
   endwin();
   exit(0);
   } /* jcn11 */

void jcn12()
   {
   int addr;
   addr = getbyte() & 0xff;
   if (rega != 0)
      pc = (pc & 0xf00) | addr;
   } /* jcn12 */

void jcn13()
   {
   mvaddstr(3,0,"Unimplemented condition: 13");
   refresh();
   endwin();
   exit(0);
   } /* jcn13 */

void jcn14()
   {
   mvaddstr(3,0,"Unimplemented condition: 14");
   refresh();
   endwin();
   exit(0);
   } /* jcn14 */

void jcn15()
   {
   mvaddstr(3,0,"Unimplemented condition: 15");
   refresh();
   endwin();
   exit(0);
   } /* jcn15 */

void jun(page)
int page;
   {
   int addr;
   addr = getbyte() & 0xff;
   pc = (page << 8) | addr;
   } /* jun */

void jms(page)
int page;
   {
   int addr;
   char str[128];
   if (regs >= 3)
      {
      sprintf(str,"Stack overflow");
      mvaddstr(3,0,str);
      refresh();
      endwin();
      exit(0);
      }
   addr = getbyte() & 0xff;
   regs++;
   stack[regs] = pc;
   pc = (page << 8) | addr;
   } /* jms */

void bbl(accum)
int accum;
   {
   int page,addr;
   char str[128];
   if (regs <= 0)
      {
      sprintf(str,"Stack underflow");
      mvaddstr(3,0,str);
      refresh();
      endwin();
      exit(0);
      }
   pc = stack[regs];
   rega = accum;
   regs--;
   } /* bbl */

void fim(hireg,loreg)
int *hireg,*loreg;
   {
   int data;
   data = getbyte() & 0xff;
   *hireg = (data >> 4) & 15;
   *loreg = data & 15;
   } /* fim */

void src(hireg,loreg)
int hireg,loreg;
   {
   src1 = hireg & 15;
   src2 = loreg & 15;
   } /* src regpair */

void fin(hireg,loreg)
int *hireg,*loreg;
   {
   int lo,hi,addr;
   unsigned char *p;
   lo = reg0 & 15;
   hi = reg1 & 15;
   addr = (pc & 0xf00) | (hi << 4) | lo;
   p = pgm + addr;
   *hireg = (*p >> 4) & 15;
   *loreg = *p & 15;
   } /* fin */

void inc(reg)
int *reg;
   {
   (*reg)++;
   if (*reg > 15) fcarry = 1;
   else fcarry = 0;
   *reg = *reg & 15;
   } /* inc reg */

void iac()
   {
   rega++;
   if (rega > 15) fcarry = 1;
   else fcarry = 0;
   rega = rega & 15;
   } /* iac */

void dac()
   {
   rega--;
   if (rega & ~15) fcarry = 1;
   else fcarry = 0;
   rega = rega & 15;
   } /* dac */

void add(reg)
int *reg;
   {
   *reg = *reg + rega + fcarry;
   if (*reg > 15) fcarry = 1;
   else fcarry = 0;
   *reg = *reg & 15;
   } /* add reg */

void sub(reg)
int *reg;
   {
   *reg = rega - *reg - fcarry;
   if (*reg & ~15) fcarry = 1;
   else fcarry = 0;
   *reg = *reg & 15;
   } /* sub reg */

void sbm()
   {
   int lo,hi,addr,mem;
   hi = src1 & 15;
   lo = src2 & 15;
   addr = (hi << 4) | lo;
   mem = ramchar[addr] & 15;
   rega = rega - mem - fcarry;
   if (rega & ~15) fcarry = 1;
   else fcarry = 0;
   rega = rega & 15;
   } /* sbm */

void adm()
   {
   int lo,hi,addr,mem;
   hi = src1 & 15;
   lo = src2 & 15;
   addr = (hi << 4) | lo;
   mem = ramchar[addr] & 15;
   rega = rega + mem + fcarry;
   if (rega & ~15) fcarry = 1;
   else fcarry = 0;
   rega = rega & 15;
   } /* adm */

void xch(reg)
int *reg;
   {
   int accum;
   accum = rega;
   rega = *reg;
   *reg = accum;
   } /* xch reg */

void isz(reg)
int *reg;
   {
   int page,addr;
   addr = getbyte() & 0xff;
   page = (pc >> 8) & 15;
   (*reg)++;
   *reg = *reg & 15;
   if (*reg != 0)
      pc = (page << 8) | addr;
   } /* isz reg,shortaddr */

void ldm(imm)
int imm;
   {
   rega = imm & 15;
   } /* ldm */

void wrm()
   {
   int lo,hi,ramaddr;
   hi = src1 & 15;
   lo = src2 & 15;
   ramaddr = (hi << 4) | lo;
   ramchar[ramaddr] = rega & 15;
   } /* wrm */

void rdm()
   {
   int lo,hi,ramaddr;
   hi = src1 & 15;
   lo = src2 & 15;
   ramaddr = (hi << 4) | lo;
   rega = ramchar[ramaddr] & 15;
   } /* rdm */

void wmp()
   {
   int hi,port;
   hi = src1 & 15;
   port = (hi >> 2) & 15;
   ramport[port] = rega & 15;
   } /* wmp */

void wrr()
   {
   int port;
   port = src1 & 15;
   romoutp[port] = rega & 15;
   } /* wrr */

void rdr()
   {
   int port;
   port = src1 & 15;
   rega = rominpt[port] & 15;
   } /* rdr */

void wrstat(stat)
int stat;
   {
   int statreg;
   statreg = stat & 15;
   ramstat[statreg] = rega & 15;
   } /* wrstat */

void rdstat(stat)
int stat;
   {
   int statreg;
   statreg = stat & 15;
   rega = ramstat[statreg] & 15;
   } /* rdstat */

void ral()
   {
   int carrybit;
   carrybit = fcarry & 1;
   rega = (rega << 1) | carrybit;
   if (rega & ~15) fcarry = 1;
   else fcarry = 0;
   rega = rega & 15;
   } /* ral */

void rar()
   {
   int carrybit;
   carrybit = (fcarry & 1) << 3;
   fcarry = rega & 1;
   rega = (rega >> 1) | carrybit;
   rega = rega & 15;
   } /* rar */

void tcc()
   {
   rega = fcarry & 1;
   fcarry = 0;
   } /* tcc */

void tcs()
   {
   rega = 9 + (fcarry & 1);
   fcarry = 0;
   } /* tcs */

void daa()
   {
   int accum;
   accum = rega & 15;
   if (accum > 9)
      {
      accum += 6;
      if (accum & ~15) fcarry = 1;
      else fcarry = 0;
      }
   rega = accum & 15;
   } /* daa */

void kbp()
   {
   int accum;
   accum = rega & 15;
   if (accum > 2)
      {
      if (accum == 4) rega = 3;
      else if (accum == 8) rega = 4;
      else rega = 15;
      }
   } /* kbp */

int main(argc,argv)
int argc;
char **argv;
   {
   int i,stat;
   unsigned int ch;
   char fname[128],str[128];
   if (argc != 2) putstx(*argv);
   strcpy(fname,*(argv+1));
   strcpy(hxtbl,"0123456789ABCDEF");
   for (i=0;i<256;i++) ramchar[i] = 0;
   for (i=0;i<16;i++)
      {
      rominpt[i] = romoutp[i] = 0;
      }
   pgm = (unsigned char *) malloc(1024*1024);
   eofsw = 0;
   loadpgm(fname);
   initnc();
   disclaim();
   while (1)
      {
      dmpreg();
      if (pc == 0xfff) break;
      ch = getch();
      ch = toupper(ch);
      if (ch == 'I')
         {
         readrom();
         continue;
         }
      else if (ch == 'G')
         {
         gotoloc();
         continue;
         }
      else if (ch == 'R')
         {
         loadpgm(fname);
         continue;
         }
      else if (ch == 'D')
         {
         dmpmem();
         continue;
         }
      else if (ch == 'M')
         {
         shwmem();
         continue;
         }
      else if (ch == 'T')
         {
         ftest ^= 1;
         continue;
         }
      else if (ch == 0x1b) break;
      ch = getbyte();
      switch (ch)
         {
         case 0x00: /* nop */
            break;
         case 0x10:
            jcn0();
            break;
         case 0x11:
            jcn1();
            break;
         case 0x12:
            jcn2();
            break;
         case 0x13:
            jcn3();
            break;
         case 0x14:
            jcn4();
            break;
         case 0x15:
            jcn5();
            break;
         case 0x16:
            jcn6();
            break;
         case 0x17:
            jcn7();
            break;
         case 0x18:
            jcn8();
            break;
         case 0x19:
            jcn9();
            break;
         case 0x1a:
            jcn10();
            break;
         case 0x1b:
            jcn11();
            break;
         case 0x1c:
            jcn12();
            break;
         case 0x1d:
            jcn13();
            break;
         case 0x1e:
            jcn14();
            break;
         case 0x1f:
            jcn15();
            break;
         case 0x20:
            fim(&reg0,&reg1);
            break;
         case 0x21:
            src(reg0,reg1);
            break;
         case 0x22:
            fim(&reg2,&reg3);
            break;
         case 0x23:
            src(reg2,reg3);
            break;
         case 0x24:
            fim(&reg4,&reg5);
            break;
         case 0x25:
            src(reg4,reg5);
            break;
         case 0x26:
            fim(&reg6,&reg7);
            break;
         case 0x27:
            src(reg6,reg7);
            break;
         case 0x28:
            fim(&reg8,&reg9);
            break;
         case 0x29:
            src(reg8,reg9);
            break;
         case 0x2a:
            fim(&reg10,&reg11);
            break;
         case 0x2b:
            src(reg10,reg11);
            break;
         case 0x2c:
            fim(&reg12,&reg13);
            break;
         case 0x2d:
            src(reg12,reg13);
            break;
         case 0x2e:
            fim(&reg14,&reg15);
            break;
         case 0x2f:
            src(reg14,reg15);
            break;
         case 0x30:
            fin(&reg0,&reg1);
            break;
         case 0x32:
            fin(&reg2,&reg3);
            break;
         case 0x34:
            fin(&reg4,&reg5);
            break;
         case 0x36:
            fin(&reg6,&reg7);
            break;
         case 0x38:
            fin(&reg8,&reg9);
            break;
         case 0x3a:
            fin(&reg10,&reg11);
            break;
         case 0x3c:
            fin(&reg12,&reg13);
            break;
         case 0x3e:
            fin(&reg14,&reg15);
            break;
         case 0x40:
            jun(0);
            break;
         case 0x41:
            jun(1);
            break;
         case 0x42:
            jun(2);
            break;
         case 0x43:
            jun(3);
            break;
         case 0x44:
            jun(4);
            break;
         case 0x45:
            jun(5);
            break;
         case 0x46:
            jun(6);
            break;
         case 0x47:
            jun(7);
            break;
         case 0x48:
            jun(8);
            break;
         case 0x49:
            jun(9);
            break;
         case 0x4a:
            jun(10);
            break;
         case 0x4b:
            jun(11);
            break;
         case 0x4c:
            jun(12);
            break;
         case 0x4d:
            jun(13);
            break;
         case 0x4e:
            jun(14);
            break;
         case 0x4f:
            jun(15);
            break;
         case 0x50:
            jms(0);
            break;
         case 0x51:
            jms(1);
            break;
         case 0x52:
            jms(2);
            break;
         case 0x53:
            jms(3);
            break;
         case 0x54:
            jms(4);
            break;
         case 0x55:
            jms(5);
            break;
         case 0x56:
            jms(6);
            break;
         case 0x57:
            jms(7);
            break;
         case 0x58:
            jms(8);
            break;
         case 0x59:
            jms(9);
            break;
         case 0x5a:
            jms(10);
            break;
         case 0x5b:
            jms(11);
            break;
         case 0x5c:
            jms(12);
            break;
         case 0x5d:
            jms(13);
            break;
         case 0x5e:
            jms(14);
            break;
         case 0x5f:
            jms(15);
            break;
         case 0x60:
            inc(&reg0);
            break;
         case 0x61:
            inc(&reg1);
            break;
         case 0x62:
            inc(&reg2);
            break;
         case 0x63:
            inc(&reg3);
            break;
         case 0x64:
            inc(&reg4);
            break;
         case 0x65:
            inc(&reg5);
            break;
         case 0x66:
            inc(&reg6);
            break;
         case 0x67:
            inc(&reg7);
            break;
         case 0x68:
            inc(&reg8);
            break;
         case 0x69:
            inc(&reg9);
            break;
         case 0x6a:
            inc(&reg10);
            break;
         case 0x6b:
            inc(&reg11);
            break;
         case 0x6c:
            inc(&reg12);
            break;
         case 0x6d:
            inc(&reg13);
            break;
         case 0x6e:
            inc(&reg14);
            break;
         case 0x6f:
            inc(&reg15);
            break;
         case 0x70:
            isz(&reg0);
            break;
         case 0x71:
            isz(&reg1);
            break;
         case 0x72:
            isz(&reg2);
            break;
         case 0x73:
            isz(&reg3);
            break;
         case 0x74:
            isz(&reg4);
            break;
         case 0x75:
            isz(&reg5);
            break;
         case 0x76:
            isz(&reg6);
            break;
         case 0x77:
            isz(&reg7);
            break;
         case 0x78:
            isz(&reg8);
            break;
         case 0x79:
            isz(&reg9);
            break;
         case 0x7a:
            isz(&reg10);
            break;
         case 0x7b:
            isz(&reg11);
            break;
         case 0x7c:
            isz(&reg12);
            break;
         case 0x7d:
            isz(&reg13);
            break;
         case 0x7e:
            isz(&reg14);
            break;
         case 0x7f:
            isz(&reg15);
            break;
         case 0x80:
            add(&reg0);
            break;
         case 0x81:
            add(&reg1);
            break;
         case 0x82:
            add(&reg2);
            break;
         case 0x83:
            add(&reg3);
            break;
         case 0x84:
            add(&reg4);
            break;
         case 0x85:
            add(&reg5);
            break;
         case 0x86:
            add(&reg6);
            break;
         case 0x87:
            add(&reg7);
            break;
         case 0x88:
            add(&reg8);
            break;
         case 0x89:
            add(&reg9);
            break;
         case 0x8a:
            add(&reg10);
            break;
         case 0x8b:
            add(&reg11);
            break;
         case 0x8c:
            add(&reg12);
            break;
         case 0x8d:
            add(&reg13);
            break;
         case 0x8e:
            add(&reg14);
            break;
         case 0x8f:
            add(&reg15);
            break;
         case 0x90:
            sub(&reg0);
            break;
         case 0x91:
            sub(&reg1);
            break;
         case 0x92:
            sub(&reg2);
            break;
         case 0x93:
            sub(&reg3);
            break;
         case 0x94:
            sub(&reg4);
            break;
         case 0x95:
            sub(&reg5);
            break;
         case 0x96:
            sub(&reg6);
            break;
         case 0x97:
            sub(&reg7);
            break;
         case 0x98:
            sub(&reg8);
            break;
         case 0x99:
            sub(&reg9);
            break;
         case 0x9a:
            sub(&reg10);
            break;
         case 0x9b:
            sub(&reg11);
            break;
         case 0x9c:
            sub(&reg12);
            break;
         case 0x9d:
            sub(&reg13);
            break;
         case 0x9e:
            sub(&reg14);
            break;
         case 0x9f:
            sub(&reg15);
            break;
         case 0xa0:
            rega = reg0;
            break;
         case 0xa1:
            rega = reg1;
            break;
         case 0xa2:
            rega = reg2;
            break;
         case 0xa3:
            rega = reg3;
            break;
         case 0xa4:
            rega = reg4;
            break;
         case 0xa5:
            rega = reg5;
            break;
         case 0xa6:
            rega = reg6;
            break;
         case 0xa7:
            rega = reg7;
            break;
         case 0xa8:
            rega = reg8;
            break;
         case 0xa9:
            rega = reg9;
            break;
         case 0xaa:
            rega = reg10;
            break;
         case 0xab:
            rega = reg11;
            break;
         case 0xac:
            rega = reg12;
            break;
         case 0xad:
            rega = reg13;
            break;
         case 0xae:
            rega = reg14;
            break;
         case 0xaf:
            rega = reg15;
            break;
         case 0xb0:
            xch(&reg0);
            break;
         case 0xb1:
            xch(&reg1);
            break;
         case 0xb2:
            xch(&reg2);
            break;
         case 0xb3:
            xch(&reg3);
            break;
         case 0xb4:
            xch(&reg4);
            break;
         case 0xb5:
            xch(&reg5);
            break;
         case 0xb6:
            xch(&reg6);
            break;
         case 0xb7:
            xch(&reg7);
            break;
         case 0xb8:
            xch(&reg8);
            break;
         case 0xb9:
            xch(&reg9);
            break;
         case 0xba:
            xch(&reg10);
            break;
         case 0xbb:
            xch(&reg11);
            break;
         case 0xbc:
            xch(&reg12);
            break;
         case 0xbd:
            xch(&reg13);
            break;
         case 0xbe:
            xch(&reg14);
            break;
         case 0xbf:
            xch(&reg15);
            break;
         case 0xc0:
            bbl(0);
            break;
         case 0xc1:
            bbl(1);
            break;
         case 0xc2:
            bbl(2);
            break;
         case 0xc3:
            bbl(3);
            break;
         case 0xc4:
            bbl(4);
            break;
         case 0xc5:
            bbl(5);
            break;
         case 0xc6:
            bbl(6);
            break;
         case 0xc7:
            bbl(7);
            break;
         case 0xc8:
            bbl(8);
            break;
         case 0xc9:
            bbl(9);
            break;
         case 0xca:
            bbl(10);
            break;
         case 0xcb:
            bbl(11);
            break;
         case 0xcc:
            bbl(12);
            break;
         case 0xcd:
            bbl(13);
            break;
         case 0xce:
            bbl(14);
            break;
         case 0xcf:
            bbl(15);
            break;
         case 0xd0:
            ldm(0);
            break;
         case 0xd1:
            ldm(1);
            break;
         case 0xd2:
            ldm(2);
            break;
         case 0xd3:
            ldm(3);
            break;
         case 0xd4:
            ldm(4);
            break;
         case 0xd5:
            ldm(5);
            break;
         case 0xd6:
            ldm(6);
            break;
         case 0xd7:
            ldm(7);
            break;
         case 0xd8:
            ldm(8);
            break;
         case 0xd9:
            ldm(9);
            break;
         case 0xda:
            ldm(10);
            break;
         case 0xdb:
            ldm(11);
            break;
         case 0xdc:
            ldm(12);
            break;
         case 0xdd:
            ldm(13);
            break;
         case 0xde:
            ldm(14);
            break;
         case 0xdf:
            ldm(15);
            break;
         case 0xe0:
            wrm();
            break;
         case 0xe1:
            wmp();
            break;
         case 0xe2:
            wrr();
            break;
         case 0xe4:
            wrstat(0);
            break;
         case 0xe5:
            wrstat(1);
            break;
         case 0xe6:
            wrstat(2);
            break;
         case 0xe7:
            wrstat(3);
            break;
         case 0xe8:
            sbm();
            break;
         case 0xe9:
            rdm();
            break;
         case 0xea:
            rdr();
            break;
         case 0xeb:
            adm();
            break;
         case 0xec:
            rdstat(0);
            break;
         case 0xed:
            rdstat(1);
            break;
         case 0xee:
            rdstat(2);
            break;
         case 0xef:
            rdstat(3);
            break;
         case 0xf0: /* clb */
            rega = fcarry = 0;
            break;
         case 0xf1: /* clc */
            fcarry = 0;
            break;
         case 0xf2:
            iac();
            break;
         case 0xf3: /* cmc */
            fcarry = fcarry ^ 1;
            break;
         case 0xf4: /* cma */
            rega = rega ^ 15;
            break;
         case 0xf5:
            ral();
            break;
         case 0xf6:
            rar();
            break;
         case 0xf7:
            tcc();
            break;
         case 0xf8:
            dac();
            break;
         case 0xf9:
            tcs();
            break;
         case 0xfa: /* stc */
            fcarry = 1;
            break;
         case 0xfb:
            daa();
            break;
         case 0xfc:
            kbp();
            break;
         case 0xfd: /* dcl */
            cml = rega & 15;
            break;
         default:
            dmpreg();
      	    sprintf(str,"%04x: Unrecognized "
      	       "instruction %02x\n", pc-1, ch&0xff);
            mvaddstr(21,0,str);
            refresh();
            dmpmem();
            endwin();
            exit(0);
         } /* switch (ch) */
      } /* instruction loop */
   endwin();
   return(0);
   } /* main */
